# hello.s
# sample RS/6000 assembly file
# derived from AIX Version 3.2 for RISC System/6000 Assembler Language
#      Reference, SC23-2197-1, pp. 4-5 to 4-7
#
# equivalent to C source code:
#  main()
#    {printf("hello, world\n");}
#
# Assemble as follows:
#  as -o hello.out -l hello.s
#

	.file "hello.s"

# static data entry in TableOfContents
	.toc
T.data:	.tc data[tc], data[rw]
	.globl	main[ds]

# main[ds] contains definitions for
# runtime linkage of function main
	.csect	main[ds]
	.long	.main[PR]
	.long	TOC[tc0]
	.long	0

# Function entry in TableOfContents
	.toc
T.hello:
	.tc .main[tc],main[ds]
	.globl	.main[PR]

# Set routine stack variables
# Values are specific to the current routine and can
# vary from routine to routine
	.set	argarea,32		# argument area
	.set	linkarea,24	# link area
	.set	locstckarea,0	# local stack storage area
	.set	ngprs,1		# # of GPR's
	.set	nfprs,0		# # of FPR;s
# total size of data storage area on stack
	.set	szdsa,8*nfprs+4*ngprs+linkarea+argarea+locstckarea

################
# main routine #
################

	.csect	.main[PR]

#### prolog ####

# save link reg in r0
	mflr	0

# save CR (if modified)

# save FPR's 14-31 (if modified)

# save GPR 13-31 (if modified)
# only r31 needs to be saved
	stm	31,-8*nfprs-4*ngprs(1)

# save LR on stack if non-leaf routine
	st	0,8(1)

# decrement stack ptr and save back chain
	stu	1,-szdsa(1)

#### function body ####

# r14 <= address of static data area
	l	14,T.data(2)

# set up parameters for the printf() call
# parameters get loaded into r3-r10
# r3 <= address of _helloworld string
	cal	3,_helloworld(14)

#call printf function
	bl	.printf[PR]
	cror	31,31,31		# nop

#### epilog ####

# restore LR
	l	0,szdsa+8(1)

# restore CR (if modified)

# restore stack ptr
	ai	1,1,szdsa

# restore GPR's
	lm	31,-8*nfprs-4*ngprs(1)

# restore FPR's

# restore LR with caller's address
	mtlr	0

# return
	brl

#external variables
	.extern	.printf[PR]

########
# data #
########

	.csect	data[rw]
	.align	2

_helloworld:
	.byte	0x68,0x65,0x6c,0x6c
	.byte	0x6f,0x2c,0x20,0x77,0x6f
	.byte	0x72,0x6c,0x64,0x0a,0x00

